/**
 * Copyright (C) 2008 Atlassian
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.atlassian.theplugin.commons.bamboo;

import com.atlassian.connector.commons.api.ConnectionCfg;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;

public final class BambooBuildInfo implements BambooBuild {
	private final Date pollingTime;
	private final ConnectionCfg server;
	private final String projectName;
	private final String planName;
	@NotNull
	private final String planKey;
	private final boolean enabled;
	@NotNull
	private final BuildStatus status;
	private final Integer number;
	@Nullable
	private final String reason;
	private final String relativeBuildDate;
	@Nullable
	private final String durationDescription;
	private final String testSummary;
	private final String commitComment;
	private final int testsPassedCount;
	private final int testsFailedCount;
	private final Throwable exception;

	private final String errorMessage;
	@Nullable
	private final Date startDate;
	private final Date completionDate;
	private final Set<String> commiters;
	// this field keeps last build status in case the current status is 'building' on 'in the queue'
	@Nullable
	private final BuildStatus lastStatus;

	public BambooBuildInfo(@NotNull String planKey, @Nullable String planName, @NotNull ConnectionCfg serverData,
			@NotNull Date pollingTime, @Nullable String projectName, boolean isEnabled, @Nullable Integer number,
			@NotNull BuildStatus status, @Nullable String reason, @Nullable Date startDate,
			@Nullable String testSummary, @Nullable String commitComment, final int testsPassedCount,
			final int testsFailedCount, @Nullable Date completionDate, @Nullable String errorMessage,
			final Throwable exception, @Nullable String relativeBuildDate, @Nullable String durationDescription,
 @Nullable Collection<String> commiters,
			@Nullable BuildStatus lastStatus) {
		this.exception = exception;
		this.lastStatus = lastStatus;
		this.pollingTime = new Date(pollingTime.getTime());
		this.planKey = planKey;
		this.planName = planName;
		this.server = serverData;
		this.projectName = projectName;
		this.enabled = isEnabled;
		this.number = number;
		this.status = status;
		this.reason = reason;
		this.testSummary = testSummary;
		this.commitComment = commitComment;
		this.testsPassedCount = testsPassedCount;
		this.testsFailedCount = testsFailedCount;
		this.errorMessage = errorMessage;
		this.relativeBuildDate = relativeBuildDate;
		this.durationDescription = durationDescription;
		this.startDate = (startDate != null) ? new Date(startDate.getTime()) : null;
		this.completionDate = (completionDate != null) ? new Date(completionDate.getTime()) : null;
		if (commiters != null) {
			this.commiters = new TreeSet<String>(commiters);
		} else {
			this.commiters = new HashSet<String>();
		}
	}

	public ConnectionCfg getServer() {
		return server;
	}

	@Nullable
	public Date getCompletionDate() {
		return completionDate == null ? null : new Date(completionDate.getTime());
	}

	public String getServerUrl() {
		return server.getUrl();
	}

	public String getBuildUrl() {
		return getServerUrl() + "/browse/" + this.planKey;
	}

	public String getResultUrl() {
		String url = getServerUrl() + "/browse/" + this.planKey;
		if (this.getStatus() != BuildStatus.UNKNOWN && this.number != null) {
			url += "-" + this.number;
		}

		return url;
	}

	public String getProjectName() {
		return projectName;
	}

	public String getPlanName() {
		return planName;
	}

	@NotNull
	public String getPlanKey() {
		return planKey;
	}

	public boolean getEnabled() {
		return enabled;
	}

	public boolean isValid() {
		return number != null;
	}

	/**
	 * @return build number
	 * @throws UnsupportedOperationException in case this object represents invalid build
	 */
	public int getNumber() throws UnsupportedOperationException {
		if (number == null) {
			throw new UnsupportedOperationException("This build has no number information");
		}
		return number;
	}

	@Nullable
	public String getReason() {
		return reason;
	}

	public String getRelativeBuildDate() {
		return relativeBuildDate;
	}

	@Nullable
	public String getDurationDescription() {
		return durationDescription;
	}

	public String getTestSummary() {
		return testSummary;
	}

	public String getCommitComment() {
		return commitComment;
	}

	@NotNull
	public BuildStatus getStatus() {
		return status;
	}

	public String getErrorMessage() {
		return this.errorMessage;
	}

	public Throwable getException() {
		return exception;
	}

	public int getTestsPassed() {
		return this.testsPassedCount;
	}

	public int getTestsFailed() {
		return this.testsFailedCount;
	}

	@Nullable
	public Date getStartDate() {
		return startDate != null ? new Date(this.startDate.getTime()) : null;
	}

	@Override
	public String toString() {
		return projectName
				+ " " + planName
				+ " " + planKey
				+ " " + status
				+ " " + reason
				+ " " + startDate
				+ " " + durationDescription
				+ " " + testSummary
				+ " " + commitComment;
	}

	/**
	 * @return whether I'm one of the commiters to that build or not
	 */
	public boolean isMyBuild() {
		return commiters.contains(server.getUsername());
	}

	/**
	 * @return list of commiters for this build
	 */
	public Set<String> getCommiters() {
		return commiters;
	}

	public Date getPollingTime() {
		return new Date(pollingTime.getTime());
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((number == null) ? 0 : number.hashCode());
		result = prime * result + ((planKey == null) ? 0 : planKey.hashCode());
		result = prime * result + ((planName == null) ? 0 : planName.hashCode());
		result = prime * result + ((projectName == null) ? 0 : projectName.hashCode());
		result = prime * result + ((server == null) ? 0 : server.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}

		BambooBuildInfo other = (BambooBuildInfo) obj;

		if (number == null) {
			if (other.number != null) {
				return false;
			}
		} else if (!number.equals(other.number)) {
			return false;
		}
		if (planKey == null) {
			if (other.planKey != null) {
				return false;
			}
		} else if (!planKey.equals(other.planKey)) {
			return false;
		}
		if (planName == null) {
			if (other.planName != null) {
				return false;
			}
		} else if (!planName.equals(other.planName)) {
			return false;
		}
		if (projectName == null) {
			if (other.projectName != null) {
				return false;
			}
		} else if (!projectName.equals(other.projectName)) {
			return false;
		}
		if (server == null) {
			if (other.server != null) {
				return false;
			}
		} else if (!server.equals(other.server)) {
			return false;
		}
		return true;
	}

	public BuildStatus getLastStatus() {
		return lastStatus;
	}

	@SuppressWarnings({"InnerClassFieldHidesOuterClassField"})
	public static class Builder {
		private final String planKey;
		private final String planName;
		private final ConnectionCfg serverData;
		private final String projectName;
		private final Integer buildNumber;
		@NotNull
		private BuildStatus buildState;
		private boolean isEnabled = true;
		private String message;
		private Date startTime;
		private Collection<String> commiters;
		private Date pollingTime = new Date();
		private String buildReason;
		@Nullable
		private String testSummary;
		@Nullable
		private String commitComment;
		private int testsPassedCount;
		private int testsFailedCount;
		private Date completionTime;
		@Nullable
		private String relativeBuildDate;
		@Nullable
		private String durationDescription;
		private Throwable exception;
		@Nullable
		private BuildStatus lastStatus;

		public Builder(@NotNull String planKey, @NotNull ConnectionCfg serverData, @NotNull BuildStatus state) {
			this.planKey = planKey;
			this.serverData = serverData;
			this.buildState = state;
			planName = null;
			projectName = null;
			buildNumber = null;
		}

		public Builder(@NotNull String planKey, @Nullable String planName, @NotNull ConnectionCfg serverData,
				@Nullable String projectName, @Nullable Integer buildNumber, @NotNull BuildStatus state) {
			this.planKey = planKey;
			this.planName = planName;
			this.serverData = serverData;
			this.projectName = projectName;
			this.buildNumber = buildNumber;
			this.buildState = state;
		}

		public Builder(BambooBuild latestBuildForPlan) {
			this.planKey = latestBuildForPlan.getPlanKey();
			this.planName = latestBuildForPlan.getPlanName();
			this.serverData = latestBuildForPlan.getServer();
			this.projectName = latestBuildForPlan.getProjectName();
			this.buildNumber = latestBuildForPlan.getNumber();
			this.buildState = latestBuildForPlan.getStatus();

			this.isEnabled = latestBuildForPlan.getEnabled();
			this.message = latestBuildForPlan.getErrorMessage();
			this.startTime = latestBuildForPlan.getStartDate();
			this.commiters = latestBuildForPlan.getCommiters();
			this.pollingTime = latestBuildForPlan.getPollingTime();
			this.buildReason = latestBuildForPlan.getReason();
			this.testSummary = latestBuildForPlan.getTestSummary();
			// this.commitComment
			this.testsPassedCount = latestBuildForPlan.getTestsPassed();
			this.testsFailedCount = latestBuildForPlan.getTestsFailed();
			this.completionTime = latestBuildForPlan.getCompletionDate();
			this.relativeBuildDate = latestBuildForPlan.getRelativeBuildDate();
			this.durationDescription = latestBuildForPlan.getDurationDescription();
			this.exception = latestBuildForPlan.getException();
			this.lastStatus = latestBuildForPlan.getLastStatus();
		}

		public Builder enabled(boolean aIsEnabled) {
			isEnabled = aIsEnabled;
			return this;
		}

		public Builder reason(String aReason) {
			this.buildReason = aReason;
			return this;
		}

		public Builder errorMessage(String aMessage) {
			this.message = aMessage;
			return this;
		}

		public Builder errorMessage(String aMessage, Throwable aException) {
			this.message = aMessage;
			this.exception = aException;
			return this;
		}

		public Builder startTime(Date aStartTime) {
			this.startTime = aStartTime;
			return this;
		}

		public Builder completionTime(Date aCompletionTime) {
			this.completionTime = aCompletionTime;
			return this;
		}


		public Builder commiters(final Collection<String> aCommiters) {
			this.commiters = aCommiters;
			return this;
		}

		public Builder testSummary(@Nullable final String aTestSummary) {
			this.testSummary = aTestSummary;
			return this;
		}

		public Builder commitComment(@Nullable final String aCommitComment) {
			this.commitComment = aCommitComment;
			return this;
		}

		public Builder pollingTime(final Date aPollingTime) {
			this.pollingTime = aPollingTime;
			return this;
		}

		public Builder testsPassedCount(int aTestsPassedCount) {
			this.testsPassedCount = aTestsPassedCount;
			return this;
		}

		public Builder testsFailedCount(int aTestsFailedCount) {
			this.testsFailedCount = aTestsFailedCount;
			return this;
		}

		public Builder relativeBuildDate(@Nullable String aRelativeBuildDate) {
			this.relativeBuildDate = aRelativeBuildDate;
			return this;
		}

		public Builder durationDescription(@Nullable String aDurationDescription) {
			this.durationDescription = aDurationDescription;
			return this;
		}

		public void buildStatus(BuildStatus buildStatus) {
			this.buildState = buildStatus;
		}

		public BambooBuildInfo build() {
			return new BambooBuildInfo(planKey, planName, serverData, pollingTime, projectName,
					isEnabled, buildNumber, buildState, buildReason, startTime, testSummary, commitComment, testsPassedCount,
 testsFailedCount,
					completionTime, message, exception, relativeBuildDate, durationDescription, commiters, lastStatus);
		}

		public void lastStatus(BuildStatus buildStatus) {
			this.lastStatus = buildStatus;
		}
	}
}
